<?php

namespace Drupal\exif_removal;

use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\File\FileSystemInterface;
use Drupal\Core\Image\ImageFactory;
use Drupal\Core\Logger\LoggerChannelFactoryInterface;
use Drupal\Core\StreamWrapper\StreamWrapperManagerInterface;
use Drupal\file\FileInterface;

/**
 * Removes sensitive EXIF data from images.
 */
class ExifRemoval {

  /**
   * The logger channel.
   *
   * @var \Drupal\Core\Logger\LoggerChannelInterface
   */
  protected $logger;

  /**
   * Inject dependencies.
   *
   * @param \Drupal\Core\File\FileSystemInterface $fileSystem
   *   File system service.
   * @param \Drupal\Core\StreamWrapper\StreamWrapperManagerInterface $streamWrapperManager
   *   Stream wrapper service.
   * @param \Drupal\Core\Image\ImageFactory $imageFactory
   *   Image factory.
   * @param \Drupal\Core\Extension\ModuleHandlerInterface $moduleHandler
   *   Module Handler.
   * @param \Drupal\Core\Logger\LoggerChannelFactoryInterface $loggerFactory
   *   The logger factory service.
   */
  public function __construct(
    protected FileSystemInterface $fileSystem,
    protected StreamWrapperManagerInterface $streamWrapperManager,
    protected ImageFactory $imageFactory,
    protected ModuleHandlerInterface $moduleHandler,
    LoggerChannelFactoryInterface $loggerFactory,
  ) {
    $this->logger = $loggerFactory->get('exif_removal');
  }

  /**
   * Removes sensitive EXIF data from images while preserving orientation.
   *
   * Removes GPS information, device details, dates, and other sensitive
   * metadata while keeping only the orientation data for proper display.
   *
   * @param \Drupal\file\FileInterface $file
   *   File entity.
   *
   * @return bool
   *   TRUE if EXIF data was successfully removed, FALSE otherwise.
   */
  public function exifRemoval(FileInterface $file) {

    if (!function_exists('exif_read_data')) {
      // EXIF functions not available.
      return FALSE;
    }

    $mime_types = ['image/jpeg'];

    if (!in_array($file->getMimeType(), $mime_types)) {
      // Not a supported mime type.
      return FALSE;
    }

    try {
      $uri = $file->getFileUri();

      // Can't get stream wrapper if URI is without scheme.
      if (!$this->streamWrapperManager::getScheme($uri)) {
        $path = $this->fileSystem->realpath($uri);
      }
      else {
        /** @var \Drupal\Core\StreamWrapper\StreamWrapperInterface $wrapper */
        $wrapper = $this->streamWrapperManager->getViaUri($uri);
        $path = $wrapper->realpath();
        if (empty($path)) {
          $path = $wrapper->getExternalUrl();
        }
      }

      if (!$path || !file_exists($path)) {
        // File path could not be determined or file doesn't exist.
        return FALSE;
      }

      $file_exif = @exif_read_data($path);

      // If no EXIF data, nothing to do.
      if (!is_array($file_exif)) {
        // No EXIF data to remove.
        return TRUE;
      }

      // Remove all EXIF data from the image.
      $image = $this->imageFactory->get($file->getFileUri());
      if ($this->moduleHandler->moduleExists('image_effects')
        && $image->getToolkit()->getPluginId() == 'imagemagick') {
        $toolkit = $image->getToolkit();
        if ($toolkit->apply('strip')) {
          $image->save();
          return TRUE;
        }
      }
      else {
        // Use GD to strip all EXIF data by re-saving the image.
        $image_resource = @imagecreatefromjpeg($path);
        if (!$image_resource) {
          // Failed to create image resource.
          return FALSE;
        }

        // Create a new image without EXIF data.
        imagejpeg($image_resource, $path, 80);
        imagedestroy($image_resource);

        return TRUE;
      }
      return TRUE;

    }
    catch (\Exception $e) {
      // Log the exception using Drupal's logger.
      $this->logger->error('Error removing EXIF data: @message for file @uri', [
        '@message' => $e->getMessage(),
        '@uri' => $file->getFileUri(),
      ]);
      return FALSE;
    }
  }

}
