<?php

declare(strict_types=1);

namespace Drupal\Tests\exif_removal\Functional;

use Drupal\Core\File\FileSystemInterface;
use Drupal\Core\Site\Settings;
use Drupal\file\Entity\File;
use Drupal\Tests\BrowserTestBase;

/**
 * @file
 * Tests for exif_removal.module.
 */

/**
 * Tests exif_removal module.
 *
 * @group exif_removal
 */
class ExifRemovalTest extends BrowserTestBase {

  /**
   * The profile to install as a basis for testing.
   *
   * Using the standard profile to test user picture config provided by the
   * standard profile.
   *
   * @var string
   */
  protected $profile = 'standard';


  /**
   * {@inheritdoc}
   */
  protected $defaultTheme = 'stark';

  /**
   * Admin user account.
   *
   * @var \Drupal\user\Entity\User
   */
  protected $adminUser;

  /**
   * Modules to enable.
   *
   * @var array
   */
  protected static $modules = ['system', 'image', 'exif_removal'];

  /**
   * The image factory service.
   *
   * @var \Drupal\Core\Image\ImageFactory
   */
  protected $imageFactory;

  /**
   * {@inheritdoc}
   */
  public static function getInfo() {
    return [
      'name' => 'EXIF Removal',
      'description' => 'Tests automatic EXIF data removal.',
      'group' => 'EXIF Removal',
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function setUp(): void {
    parent::setUp();

    // Create admin user with all necessary permissions.
    $this->adminUser = $this->drupalCreateUser([
      'administer site configuration',
      'access content',
      'access administration pages',
      'administer users',
      'administer account settings',
      'access user profiles',
    ]);
    $this->drupalLogin($this->adminUser);

    // Test if directories specified in settings exist in filesystem.
    $file_dir = Settings::get('file_public_path');
    \Drupal::service('file_system')->prepareDirectory($file_dir, FileSystemInterface::CREATE_DIRECTORY);

    $picture_dir = \Drupal::state()->get('user_picture_path', 'pictures');
    $picture_path = $file_dir . '/' . $picture_dir;

    \Drupal::service('file_system')->prepareDirectory($picture_path, FileSystemInterface::CREATE_DIRECTORY);
    $directory_writable = is_writable($picture_path);
    $this->assertTrue($directory_writable, "The directory $picture_path doesn't exist or is not writable. Further tests won't be made.");

    $this->imageFactory = $this->container->get('image.factory');

    // Verify user picture field is enabled.
    $this->assertTrue($this->config('field.field.user.user.user_picture')->get('status'), 'User picture field is enabled');
  }

  /**
   * Tests Image toolkit setup form.
   */
  public function testToolkitSetupForm() {
    // Get form.
    $this->drupalGet('admin/config/media/image-toolkit');

    // Test that default toolkit is GD.
    $this->assertSession()->fieldValueEquals('image_toolkit', 'gd');

    // Test changing the jpeg image quality.
    $edit = ['gd[image_jpeg_quality]' => '98'];
    $this->submitForm($edit, 'Save configuration');
    $this->assertEquals('98', $this->config('system.image.gd')->get('jpeg_quality'));
  }

  /**
   * Test EXIF data removal of uploaded user profile pictures.
   */
  public function testUserPicture() {
    // Ensure we're logged in.
    $this->drupalLogin($this->adminUser);
    $this->assertSession()->statusCodeEquals(200);

    // Verify user picture field configuration.
    $field_config = $this->config('field.field.user.user.user_picture');
    $this->assertTrue($field_config->get('status'), 'User picture field is enabled');

    // Verify file directory exists and is writable.
    $file_dir = Settings::get('file_public_path');
    $picture_dir = \Drupal::state()->get('user_picture_path', 'pictures');
    $picture_path = $file_dir . '/' . $picture_dir;
    $this->assertTrue(is_dir($picture_path), "Directory exists: $picture_path");
    $this->assertTrue(is_writable($picture_path), "Directory is writable: $picture_path");

    // Create a test file directly to verify file operations work.
    $test_content = 'Test file content';
    $test_file = $picture_path . '/test_file.txt';
    $this->assertTrue(file_put_contents($test_file, $test_content) !== FALSE, 'Test file created successfully');
    $this->assertTrue(file_exists($test_file), 'Test file exists');
    $this->assertEquals($test_content, file_get_contents($test_file), 'Test file content matches');

    // Upload user picture.
    $file = $this->saveUserPicture();

    // Verify file was created and EXIF data was removed.
    $this->assertNotNull($file, 'File entity was created');
    if ($file) {
      $this->assertExifDataRemoved($file);
    }
  }

  /**
   * Uploads a user picture.
   */
  private function saveUserPicture() {
    $test_file_path = \Drupal::service('extension.list.module')->getPath('exif_removal') . '/tests/exif_example.jpg';

    // Verify test file exists.
    $file_system = \Drupal::service('file_system');
    $real_path = $file_system->realpath($test_file_path);
    $this->assertNotEmpty($real_path, 'Test file exists at: ' . $test_file_path);
    $this->assertTrue(file_exists($real_path), 'Test file is accessible');

    // Check file size and readability.
    $this->assertTrue(filesize($real_path) > 0, 'Test file has content');
    $this->assertTrue(is_readable($real_path), 'Test file is readable');

    // Get the user edit form.
    $uid = $this->adminUser->id();
    $this->drupalGet('user/' . $uid . '/edit');
    $this->assertSession()->statusCodeEquals(200);

    // Verify the form has the file upload field.
    $this->assertSession()->fieldExists('files[user_picture_0]');

    // Submit the form with the file.
    $edit = ['files[user_picture_0]' => $real_path];
    $this->submitForm($edit, 'Save');

    // Verify form submission was successful.
    $this->assertSession()->pageTextContains('The changes have been saved.');

    // Check for any form errors.
    $errors = $this->getSession()->getPage()->findAll('css', '.error-message');
    if (!empty($errors)) {
      foreach ($errors as $error) {
        $this->assertTrue(FALSE, 'Form error: ' . $error->getText());
      }
    }

    // Load actual user data from database.
    $user_storage = $this->container->get('entity_type.manager')->getStorage('user');
    $user_storage->resetCache([$this->adminUser->id()]);

    $account = $user_storage->load($this->adminUser->id());

    // Debug account fields.
    $this->assertNotNull($account, 'User account loaded');
    $this->assertEquals($uid, $account->id(), 'User ID matches');

    // Check if user_picture field exists on the account.
    $this->assertTrue($account->hasField('user_picture'), 'User picture field exists on account');

    // Debug user picture field.
    if ($account->hasField('user_picture')) {
      $picture_field = $account->get('user_picture');
      $this->assertNotNull($picture_field, 'User picture field is not null');
      $this->assertNotEmpty($picture_field, 'User picture field is not empty');

      if (!empty($picture_field)) {
        $target_id = $picture_field->target_id;
        $this->assertNotEmpty($target_id, 'User picture target_id is not empty');

        // Load and return the file entity.
        $file = File::load($target_id);
        $this->assertNotEmpty($file, 'File entity loaded successfully');

        if ($file) {
          // Debug file entity.
          $this->assertEquals($target_id, $file->id(), 'File ID matches target_id');
          $this->assertNotEmpty($file->getFileUri(), 'File URI is not empty');

          return $file;
        }
      }
    }

    return NULL;
  }

  /**
   * Verify that sensitive EXIF data has been removed from the image.
   */
  private function assertExifDataRemoved($file) {
    // Ensure we have a valid file entity.
    $this->assertNotNull($file, 'File entity is not null');
    if (!$file) {
      return;
    }

    // Get the file path.
    $file_system = \Drupal::service('file_system');
    $uri = $file->getFileUri();
    $this->assertNotEmpty($uri, 'File URI is not empty');

    $path = $file_system->realpath($uri);
    $this->assertNotEmpty($path, 'File path is not empty');
    $this->assertTrue(file_exists($path), 'File exists at path: ' . $path);

    // Read EXIF data.
    $exif_data = @exif_read_data($path);

    // If exif_read_data returns false, it means there's no EXIF data
    // which is what we want, or there was an error reading the file.
    if ($exif_data === FALSE) {
      $this->assertTrue(TRUE, 'No EXIF data found in the image (or unable to read EXIF data).');
      return;
    }

    // Check that sensitive EXIF data is not present.
    $this->assertFalse(isset($exif_data['GPS']), 'GPS data has been removed.');
    $this->assertFalse(isset($exif_data['Make']), 'Device make has been removed.');
    $this->assertFalse(isset($exif_data['Model']), 'Device model has been removed.');
    $this->assertFalse(isset($exif_data['DateTime']), 'DateTime has been removed.');
  }

}
